
##################################################
#  Python SHADE Algorithm implemantion - Based on original C++ code from Ryoji Tanabe.
#  Carlos Andres Ramirez (email: andres@antai-group.com  URL: https://carlos.engineer/ ) 
#  Last Modified on Oct. 2015
##################################################

import sys
import random
import math

class Shade:
    """Shade algorithm implementation."""

    def __init__(self):

        self.initializeParameters()

        # Create the random vectors (Population) and update the objective function (Quality)
        self.Xpopulation = list(map(lambda x: self.generate_random_individual(), range(self.np)))
        self.Xfitness = list(map(lambda i: self.func(self.Xpopulation[i]), range(0, self.np)))
        self.set_best()
        
        self.Vpopulation = []              # Children population
        self.Vfitness = []                 # Children's quality vector

        self.Apopulation = []              # Archive population
        self.arc_size   = self.np          # Archive size
        
        self.max_evaluations = self.dimensions * 10000
        
        ### mios
        self.current_generation_count = 0
        self.current_individual_count = 0

        # JADE
        #self.ucr = 0.5
        #self.uf = 0.5
        self.pop_cr = list(map(lambda i: 0, range(0, self.np)))
        self.pop_f = list(map(lambda i: 0, range(0, self.np)))
        self.p_min = 2.0 / self.np
        
        # Shade
        self.dif_fitness = []
        self.sum_dif_fitness = 0
        self.success_sf = []
        self.success_cr = []
        self.memory_pos = 0
        self.memory_cr = list(map(lambda i: 0.5, range(0, self.memory_size)))
        self.memory_sf = list(map(lambda i: 0.5, range(0, self.memory_size)))

    def initializeParameters(self):
        self.np = 15                    #
        self.total_generations = 50
        self.dimensions = 5
        self.bound_upper = 100
        self.bound_lower = -100
        
        self.f_count = 0                # Fitness count
        self.memory_size = 15           # Memory size

    def generate_random_individual(self):
        individual = list(map(lambda i: random.uniform(self.bound_lower, self.bound_upper), range(0, self.dimensions)))
        return individual

    def set_best(self):
        self.best_fitness = self.Xfitness[0]
        self.best = self.Xpopulation[0]
        for i in range(1, self.np):
             cost = self.Xfitness[0]
             if cost < self.best_fitness :
                 self.best_fitness = cost
                 self.best = self.Xpopulation[i]
    
    def func(self, individual):
        suma = sum(list(map(lambda i: individual[i] * individual[i], range(0, self.dimensions))))
        self.f_count += 1
        return suma

    def parent_to_archive(self, parent_index):
        if len(self.Apopulation) < self.arc_size:
            self.Apopulation.append(self.Xpopulation[parent_index])
        else: # Archive is full, overwrite random vector
            self.Apopulation[random.randint(0, self.arc_size - 1)] = self.Xpopulation[parent_index]

    def get_random_population_index(self):
        return random.randint(0, self.np - 1)

    def main_loop(self):
               
        while self.f_count < self.max_evaluations:
            xbest_population = self.get_pbest()

            self.Vpopulation = []   # Children list
            self.current_individual_count = 0
            while self.current_individual_count < self.np:

                #In each generation, CR_i and F_i used by each individual x_i are generated by first selecting an index r_i randomly from [1, H]
                random_selected_period = random.randint(0, self.memory_size - 1)
                mu_sf = self.memory_sf[random_selected_period]
                mu_cr = self.memory_cr[random_selected_period]
                
                #generate CR_i and repair its value
                self.pop_cr[self.current_individual_count] = self.normal_rvalue(mu_cr, 0.1)
                if self.pop_cr[self.current_individual_count] > 1:
                    self.pop_cr[self.current_individual_count] = 1
                elif self.pop_cr[self.current_individual_count] < 0:
                    self.pop_cr[self.current_individual_count] = 0

                #generate F_i and repair its value                
                self.pop_f[self.current_individual_count] = self.cauchy_rvalue(mu_sf, 0.1)
                while self.pop_f[self.current_individual_count] <= 0:
                    self.pop_f[self.current_individual_count] = self.cauchy_rvalue(mu_sf, 0.1)
                if self.pop_f[self.current_individual_count] > 1:
                    self.pop_f[self.current_individual_count] = 1

                # generate p_i in current/to/pbest/1 randomly
                p_var = (0.2 - self.p_min) * random.random() + self.p_min
                p_num = int(self.np * p_var)
                #p-best individual is randomly selected from the top np * p_i members
                pbest = self.Xpopulation[xbest_population[int(random.random() % p_num)][0]]

                #####
                # Mutation & Crossover
                #####
                tmp = self.crossover_ctpb1bin(self.Xpopulation[self.current_individual_count], pbest)
                self.Vpopulation.append(tmp)    # Add to children population
                self.current_individual_count += 1

            # Evaluate the children's fitness values and check if there is new best
            self.Vfitness = list(map(lambda i: self.func(self.Vpopulation[i]), range(0, self.np)))

            #####
            # TODO: Check CEC 2013 rules - epsilon etc.
            #####
            for i in range(0, self.np) :
                if self.Vfitness[i] < self.best_fitness :
                     print(self.Vfitness[i])
                     self.best_fitness = self.Vfitness[i]
                     self.best = self.Vpopulation[i]

            if self.f_count >= self.max_evaluations :
                break
            
            num_success_param = 0
            for i in range(0, self.np) :
                if self.Vfitness[i] == self.Xfitness[i] :
                    self.Xfitness[i] = self.Vfitness[i]
                    self.Xpopulation[i] = self.Vpopulation[i]
                elif self.Vfitness[i] < self.Xfitness[i] :
                    #parent vectors x_i which were worse than the trial vectors u_i are preserved
                    if self.arc_size > 1 :
                        self.parent_to_archive(i)
                    
                    self.dif_fitness.append(abs(self.Xfitness[i] - self.Vfitness[i]))
                    self.Xfitness[i] = self.Vfitness[i]
                    self.Xpopulation[i] = self.Vpopulation[i]
                    #successful parameters are preserved in S_F and S_CR
                    self.success_sf.append(self.pop_f[i])
                    self.success_cr.append(self.pop_cr[i])
                    num_success_param += 1

            # if numeber of successful parameters > 0, historical memories are updated 
            if num_success_param > 0 :
                self.memory_sf[self.memory_pos] = 0
                self.memory_cr[self.memory_pos] = 0
                sum_dif_fitness = 0
                temp_sum_sf = 0

                for i in range(0, num_success_param) :
                    sum_dif_fitness += self.dif_fitness[i];

                for i in range(0, num_success_param) :
                    weight = self.dif_fitness[i] / sum_dif_fitness
                    #weighted lehmer mean
                    self.memory_sf[self.memory_pos] += weight * self.success_sf[i] * self.success_sf[i]
                    temp_sum_sf += weight * self.success_sf[i]
                    #weighted arithmetic mean
                    self.memory_cr[self.memory_pos] += weight * self.success_cr[i]

                self.memory_sf[self.memory_pos] /= temp_sum_sf;

                #increment the counter
                self.memory_pos += 1
                if self.memory_pos >= self.memory_size :
                    self.memory_pos = 0

                # clear out the S_F, S_CR and delta fitness
                del self.success_sf[:]
                del self.success_cr[:]
                del self.dif_fitness[:]

            self.current_generation_count += 1
                        
        ##########
        # Program ends
        ##########
        print('Best f: ', self.best_fitness)
        print('Best: ', self.best)
        print('f count: ', self.f_count)

    def get_pbest(self):
        pbest_pop = list(map(lambda i: [i, self.Xfitness[i]], range(0, self.np)))
        pbest_pop.sort(key=lambda x: x[1])
        return pbest_pop

    def crossover_ctpb1bin(self, v1, pbest_vec):
        r1 = self.get_random_population_index()
        while r1 == self.current_individual_count :
            r1 = self.get_random_population_index()

        xapoprange = self.np + len(self.Apopulation) - 1
        r2 = random.randint(0, xapoprange)
        while r2 == self.current_individual_count or r2 == r1 :
            r2 = random.randint(0, xapoprange)
        if r2 >= self.np :
            r2vec = self.Apopulation[r2 - self.np]
        else :
            r2vec = self.Xpopulation[r2]

        trial = list(v1)
        j = random.randint(0, self.dimensions - 1)
        for i in range(0, self.dimensions) :
            if i == j or random.random() < self.pop_cr[self.current_individual_count] :
                trial[i] = v1[i] + self.pop_f[self.current_individual_count] * (pbest_vec[i] - v1[i]) + self.pop_f[self.current_individual_count] * (self.Xpopulation[r1][i] - r2vec[i])
            else :
                trial[i] = v1[i]
        return trial

    def cauchy_rvalue(self, mu, gamma):
        # Start with a uniform random sample from the open interval (0, 1).
        # But random() returns a sample from the half-open interval [0, 1).
        # In the unlikely event that random() returns 0, try again.
        p = 0.0
        while p == 0.0:
            p = random.random()
        return mu + gamma * math.tan(math.pi * (p - 0.5))

    def normal_rvalue(self, mu, sigma):
        return random.gauss(mu, sigma)

##################################################
# Main program
##################################################
myde = Shade()
myde.main_loop()
##################################################
# EOF
##################################################

